"""
This module is outdated and should not be used. An error is thrown when importing it but if you still want to use it
simply do a try/except block.
"""

from ffxivcalc.helperCode.exceptions import outDatedAPICode
raise outDatedAPICode
